/*
 * $Id: NameserverClient.java 364 2007-04-09 03:52:26Z davidb $
 *
 * Copyright (c) 2004 USC Project SRO. All rights reserved.
 * Copyright (c) 2005 VeriSign, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

package se.rfc.unbound.nsclient;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.PortUnreachableException;
import java.net.SocketTimeoutException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Properties;
import java.util.Random;
import java.util.Set;

import org.apache.log4j.Logger;
import org.xbill.DNS.*;

import se.rfc.unbound.*;
import se.rfc.unbound.nsclient.RequestMap.RequestEntry;
import se.rfc.unbound.server.Server;

/**
 * The Client module of the unbound architecture.
 * 
 * This module is responsible for converting requests to wire line format,
 * (optionally) sending the DNS request and receiving the response, parsing
 * response messages, assigning IDs to outgoing request messages, handling UDP
 * retry, fallback to TCP, EDNS negotiation, and handling TSIG application.
 * 
 * This module is always at the end of a chain of Resolver modules.
 * 
 * This module can be configured into one of two modes: synchronous and
 * asynchronous. In synchronous mode, this module will take charge of sending
 * the request, waiting for the response, and (possibly) retrying over TCP. In
 * asynchronous mode, this module will just register the request with the
 * server module, letting the server module actually handle all I/O. However,
 * it is also responsible for matching responses to active requests.
 * 
 * @author davidb
 * @version $Revision: 364 $
 */
public class NameserverClient extends AbstractResolver
{
  /** The default query timeout, in milliseconds. */
  private long            mDefaultQueryTimeout;

  /** If this is non-null, then this module is asynchronous. */
  private Server          mServer;

  /**
   * In ansynchronous mode, this module is responsible for tracking outgoing
   * requests and matching them with responses coming from the server.
   */
  private RequestMap      mRequestMap;

  /**
   * The ServerSelection module, if present.
   */
  private ServerSelection mServerSelection;

  /**
   * A source of randomness for generating new IDs. (not used in synchronous
   * mode, since DNSjava does this for us).
   */
  private static Random   mRandom = new Random();

  // this is the normal logger.
  private Logger          log     = Logger.getLogger(this.getClass());

  // this is the specific logger used for logging DNS traffic
  // private Logger trafLog = Logger.getLogger("client_traffic_log");

  /**
   * Create an asynchronous-mode NameserverClient.
   * 
   * @param server The module to register the write event to.
   */
  public NameserverClient(Server server, ServerSelection serverSelection)
  {
    // NOTE: there is no next resolver, as this one is the end of the chain.
    super(null);

    mServer = server;
    mServerSelection = serverSelection;
    mRequestMap = new RequestMap();
  }

  /**
   * Create a synchronous-mode NameserverClient.
   */
  public NameserverClient()
  {
    // NOTE: there is no next resolver, as this one is the end of the chain.
    super(null);
  }

  public void init(Properties config)
  {
    if (config == null) return;

    String s = config.getProperty("dns.client.timeout");
    mDefaultQueryTimeout = Util.parseLong(s, 0);
  }

  /**
   * Generates a random number suitable for use as a message ID
   */
  protected static int randomID()
  {
    return (mRandom.nextInt(0xffff));
  }

  /**
   * This is the synchronous method for sending a request and receiving a
   * response. This method will block until it gets a response or times out.
   * It is not possible (currently) to specify a source port or address to
   * send the request from.
   * 
   * @param event The event to be processed. In this case, the response
   *          received (if any) is attached to this event.
   * @param requestMessage The request to send.
   * @param mstate The match state contains the target to send to.
   */

  private void send(DNSEvent event, Message requestMessage, MatchState mstate)
  {
    Host target = null;
    try
    {
      // Note that this technique is ultimately limited wrt process model.
      // Get a new SimpleResolver every time -- this makes sure we get one for
      // each target.

      target = mstate.getTarget();
      String hostAddr = target.getAddress().getHostAddress();

      // Create a new resolver pointing to our current target.
      SimpleResolver res = new SimpleResolver(hostAddr);
      res.setPort(target.getPort());
      res.setTimeout((int) mstate.getTimeoutDuration() / 1000);
      res.setTCP(mstate.getUseTCP());

      long sentTime = System.currentTimeMillis();

      Message resp = res.send(requestMessage);

      long rtt = System.currentTimeMillis() - sentTime;
      if (mServerSelection != null)
      {
        mServerSelection.setRTT(target, rtt);
      }

      Response response = new Response(new SMessage(resp), Response.OK);
      event.setResponse(response);
    }
    catch (SocketTimeoutException e)
    {
      log.debug("query timed out");
      event.setResponse(Response.timeoutResponse());
    }
    catch (PortUnreachableException e)
    {
      log.warn("query failed: Port Unreachable ("
          + target.getAddress().getHostAddress() + ":" + target.getPort()
          + ")");
      event.setResponse(Response.timeoutResponse());
    }
    catch (IOException e)
    {
      log.warn("query failed: " + e.getMessage());
      // for now, we treat this as a timeout.
      event.setResponse(Response.timeoutResponse());
    }
  }

  public void processRequest(DNSEvent event)
  {
    // Get a local copy of the request.
    Request local_request = (Request) event.getRequest().clone();

    // do the "one time" processing for this particular request at this
    // particular time.

    Host target = local_request.getTarget();

    if (target.getEDNSVersion() < 0 && local_request.hasOPT())
    {
      log.trace("pre-emptively removing EDNS for this target.");
      local_request.setOPT(null);
    }

    MatchState mstate = new MatchState();
    mstate.setTimeoutDuration(mDefaultQueryTimeout);
    mstate.setWakeupDuration(local_request.getWakeupTime());

    handleRequest(event, local_request, mstate);
  }

  /**
   * Handle a request. This will be called each time a request is sent or
   * resent (to the same target).
   * 
   * @param event The request event.
   * @param state The state for the event. This should already have the
   *          volatile event information copied over.
   */
  private void handleRequest(DNSEvent event, Request request,
      MatchState mstate)
  {
    // Set a new (hopefully unique) ID for each different write request.
    request.setID(randomID());

    // Update our mstate to match the new request information.
    mstate.setRequestInfo(request);

    Message m = request.getMessage();
    mstate.setRequestMessage(m); // save this so it can be logged later.

    // TSIG should be applied here, if requested.

    // No reference to the server module means that we are in synchronous
    // mode.
    if (mServer == null)
    {
      // This will block until either a response is received or the request
      // times out.

      send(event, m, mstate);
      // Immediately process the response.
      processResponse(event);
      return;
    }

    // Conver the current request to wire format.
    byte[] out = m.toWire(512);

    Host target = mstate.getTarget();

    // And schedule it for writing.
    log.trace("scheduling request " + request + " for writing to: " + target);
    mServer.writeRequest(out,
        target.getSocketAddress(),
        event,
        mstate,
        mstate.getUseTCP()
        // true
    );
    mServer.logRequest(m, target.getSocketAddress());
  }

  /**
   * Once a request is asynchronously written, this method must be called. It
   * handles tracking the request.
   * 
   * @param event The request event.
   * @param matchState The match state that goes with this event.
   */
  public void registerWrite(DNSEvent event, Object matchState)
  {
    MatchState mstate = (MatchState) matchState;

    // Note the write time.
    mstate.setSentTime(System.currentTimeMillis());

    log.trace("wrote request to " + mstate.getTarget().getSocketAddress()
        + ":\n" + mstate.getRequestMessage());

    // Add this to the request map, so we can correlate it with responses.
    mRequestMap.addRequestEvent(event, mstate);
  }

  /**
   * When the server module reads a response that it accepts, it passes that
   * response to this routine for all futher processing. This routine is
   * responsible for correlating the response with a request, and sending the
   * combined event back up through the resolver chain.
   * 
   * @param responseMessage The response message itself, already parsed.
   * @param sourceAddress The source address and port that the message came
   *          from.
   * @return true if the response was accepted.
   */
  public boolean handleResponse(Message responseMessage,
      InetSocketAddress sourceAddress)
  {
    SMessage m = new SMessage(responseMessage);

    // attempt to correlate the response with a request.

    RequestEntry re = mRequestMap.findRequestEvent(m, sourceAddress);
    if (re == null)
    {
      log.trace("response from " + sourceAddress
          + " failed to match an existing request");
      return false;
    }

    // We matched against an actual request.

    DNSEvent event = re.event;
    MatchState mstate = re.mstate;
    Host target = mstate.getTarget();

    long rtt = System.currentTimeMillis() - mstate.getSentTime();
    if (mServerSelection != null)
    {
      mServerSelection.setRTT(target, rtt);
    }

    // Create a local copy of the current event request, set it to match the
    // mstate.
    Request local_request = (Request) event.getRequest().clone();

    local_request.setID(mstate.getRequestID());
    local_request.setQName(mstate.getRequestQuestion().getName());
    local_request.setTarget(mstate.getTarget());

    // Now check for retry conditions.

    if (m.getRcode() == Rcode.FORMERR && !mstate.getSentWithoutEDNS())
    {
      // first note that the target apparently cannot handle EDNS.
      target.setEDNSVersion(-1);
      local_request.setOPT(null);

      handleRequest(event, local_request, mstate);
      return true;
    }

    // If the response was truncated, just discard it and retry via TCP.
    if (m.getHeader().getFlag(Flags.TC) && !mstate.getUseTCP())
    {
      mstate.setUseTCP(true);

      handleRequest(event, local_request, mstate);
      return true;
    }

    // If the request doesn't need to be retried, sent it up the chain for
    // processing.
    processResponse(event);
    return true;
  }

  /**
   * Look at the set of outstanding sent requests for requests that have
   * either timed out, or are old enough to trigger a wakeup event.
   * 
   * @return The amount of time to block until the next timeout/wakeup event.
   */
  public long handleTimeouts()
  {
    List events = mRequestMap.expiredEvents();
    for (Iterator i = events.iterator(); i.hasNext();)
    {
      RequestEntry re = (RequestEntry) i.next();

      DNSEvent event = re.event;
      Response response = event.getResponse();

      // TODO: do any special wakeup or timeout processing here.
      if (response.getResponseStatus() == Response.WAKEUP)
      {
        log.trace("woke up request: " + re.mstate.getRequestID() + " to "
            + re.mstate.getTarget());
      }
      // Send the response back up the resolver chain.
      processResponse(event);
    }

    return mRequestMap.getNextTimeout(System.currentTimeMillis());
  }

  public void processResponse(DNSEvent event)
  {
    Response response = event.getResponse();

    if (response.getResponseStatus() == Response.OK)
    {
      // Normalize responses past this point.
      normalize(response.getSMessage());
    }

    mPrevResolver.processResponse(event);
  }

  /**
   * This routine normalizes a response. This includes removing "irrelevant"
   * records from the answer and additional sections and (re)synthesizing
   * CNAMEs from DNAMEs, if present.
   * 
   * @param response
   */
  private SMessage normalize(SMessage m)
  {
    if (m == null) return m;

    if (m.getRcode() != Rcode.NOERROR && m.getRcode() != Rcode.NXDOMAIN)
    {
      return m;
    }

    Name qname = m.getQuestion().getName();
    int qtype = m.getQuestion().getType();

    Name sname = qname;

    // For the ANSWER section, remove all "irrelevant" records and add
    // synthesized CNAMEs from DNAMEs
    // This will strip out-of-order CNAMEs as well.
    List rrset_list = m.getSectionList(Section.ANSWER);
    Set additional_names = new HashSet();

    for (ListIterator i = rrset_list.listIterator(); i.hasNext();)
    {
      SRRset rrset = (SRRset) i.next();
      int type = rrset.getType();
      Name n = rrset.getName();

      // Handle DNAME synthesis; DNAME synthesis does not occur at the DNAME
      // name itself.
      if (type == Type.DNAME && sname.strictSubdomain(n))
      {
        if (rrset.size() > 1)
        {
          log.debug("Found DNAME rrset with size > 1: " + rrset);
          return Util.errorMessage(m, Rcode.SERVFAIL);
        }
        DNAMERecord dname = (DNAMERecord) rrset.first();
        try
        {
          Name cname_alias = sname.fromDNAME(dname);
          // Note that synthesized CNAMEs should have a TTL of zero.

          // FIXME: resolve the conflict between synthesized CNAME ttls and
          // the cache.
          CNAMERecord cname = new CNAMERecord(sname, dname.getDClass(), 0,
              cname_alias);
          SRRset cname_rrset = new SRRset();
          cname_rrset.addRR(cname);
          i.add(cname_rrset);

          sname = cname_alias;
        }
        catch (NameTooLongException e)
        {
          log.debug("not adding synthesized CNAME -- "
              + "generated name is too long", e);
        }
        continue;
      }

      // The only records in the ANSWER section not allowed to
      if (!n.equals(sname))
      {
        log.debug("normalize: removing irrelevant rrset: " + rrset);
        i.remove();
        continue;
      }

      // Follow the CNAME chain.
      if (type == Type.CNAME)
      {
        if (rrset.size() > 1)
        {
          log.debug("Found CNAME rrset with size > 1: " + rrset);
          return Util.errorMessage(m, Rcode.SERVFAIL);
        }

        CNAMERecord cname = (CNAMERecord) rrset.first();
        sname = cname.getAlias();
        continue;
      }

      // Otherwise, make sure that the RRset matches the qtype.
      if (qtype != Type.ANY && qtype != type)
      {
        log.debug("normalize: removing irrelevant rrset: " + rrset);
        i.remove();
      }

      // Otherwise, fetch the additional names from the relevant rrset.
      rrsetAdditionalNames(additional_names, rrset);
    }

    // Get additional names from AUTHORITY
    rrset_list = m.getSectionList(Section.AUTHORITY);
    for (Iterator i = rrset_list.iterator(); i.hasNext();)
    {
      SRRset rrset = (SRRset) i.next();
      rrsetAdditionalNames(additional_names, rrset);
    }

    // For each record in the additional section, remove it if it is an
    // address record and not in the collection of additional names found in
    // ANSWER and AUTHORITY.
    rrset_list = m.getSectionList(Section.ADDITIONAL);
    for (Iterator i = rrset_list.iterator(); i.hasNext();)
    {
      SRRset rrset = (SRRset) i.next();
      int type = rrset.getType();
      if ((type == Type.A || type == Type.AAAA)
          && !additional_names.contains(rrset.getName()))
      {
        i.remove();
      }
      // FIXME: what about other types?
    }

    return m;
  }

  /**
   * Extract additional names from the records in an rrset.
   * 
   * @param additional_names The set to add the additional names to, if any.
   * @param rrset The rrset to extract from.
   */
  private void rrsetAdditionalNames(Set additional_names, RRset rrset)
  {
    if (rrset == null) return;

    for (Iterator i = rrset.rrs(); i.hasNext();)
    {
      Record r = (Record) i.next();
      Name add_name = r.getAdditionalName();
      if (add_name != null)
      {
        additional_names.add(add_name);
      }
    }
  }

}